import os
from pathlib import Path
from typing import Tuple, Callable

import networkx as nx
import numpy as np
from pydrake.all import (
    RotationMatrix, Quaternion, RollPitchYaw,
    Solve, SolutionResult,
    InverseKinematics, ModelInstanceIndex,
)
from pydrake.common.value import AbstractValue
from pydrake.math import RigidTransform
from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import Diagram, Context
from pydrake.systems.primitives import ConstantVectorSource, ConstantValueSource

# Internal Imports
from brom_drake.motion_planning.systems.prototypical_planner import PrototypicalPlannerSystem
from brom_drake.scenes.types import BaseScene
from brom_drake.scenes.roles import kKinematicMotionPlanner, Role
from brom_drake.file_manipulation.urdf.shapes.sphere import SphereDefinition
from brom_drake.file_manipulation.urdf.simple_writer.urdf_definition import SimpleShapeURDFDefinition
from brom_drake.utils import Performer, MotionPlan


class KinematicMotionPlanningScene(BaseScene):
    def __init__(
        self,
        start_configuration: np.ndarray = None,
        start_pose: RigidTransform = None,
        goal_configuration: np.ndarray = None,
        goal_pose: RigidTransform = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        # Start and Goal Configurations
        self.start_config_ = start_configuration
        self.goal_config_ = goal_configuration

        # Start and Goal Poses
        self.start_pose_ = start_pose
        self.goal_pose_ = goal_pose

        # Create placeholder for the plant
        self.plant = None

        # Create placeholder for the robot model index
        self.robot_model_idx_ = None

    def add_all_secondary_cast_members_to_builder(self):
        """
        Description
        -----------
        This method will add the start and goal poses to the builder.
        :return:
        """
        # Add visual elements for the start and goal poses
        self.add_start_and_goal_to_plant(self.plant)

    def add_robot_source_system(self):
        """
        Description
        -----------
        This method adds a source for providing the motion planner
        with the model index for the robot that we are trying to control.
        :return:
        """
        # Setup

        # Create AbstractValueSource
        robot_source_system = ConstantValueSource(
            AbstractValue.Make(self.robot_model_idx_)
        )
        robot_source_system.set_name("robot_model_index_source")

        self.builder.AddSystem(robot_source_system)

    def add_start_and_goal_to_plant(self, plant: MultibodyPlant):
        """
        Description
        -----------
        Add the start and goal to this plant.
        Parameters
        ----------
        :param plant: The plant to add the start and goal to.

        Notes
        -----
        Feel free to overwrite this if you have specific sizes of the start and goal in mind,
        or if you want to add more than just spheres, etc.
        :return:

        """

        # Create a sphere for the start
        start_sphere_defn = SimpleShapeURDFDefinition(
            name="start_sphere",
            shape=SphereDefinition(radius=0.1),
            create_collision=False,
            color=np.array([0, 0.3, 0.8, 0.3]),
        )
        start_sphere_urdf_location = Path("brom/models/spheres/start.urdf")
        os.makedirs(start_sphere_urdf_location.parent, exist_ok=True)

        start_sphere_defn.write_to_file(start_sphere_urdf_location)

        # Load the start sphere into the plant and rigidly attach it at the start_pose
        start_sphere_model_idx = Parser(plant).AddModels(str(start_sphere_urdf_location))[0]
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("start_sphere_base_link", start_sphere_model_idx),
            self.start_pose,
        )

        # Create a sphere for the goal
        goal_sphere_defn = SimpleShapeURDFDefinition(
            name="end_sphere",
            shape=SphereDefinition(radius=0.1),
            create_collision=False,
            color=np.array([0, 0.8, 0.3, 0.3]),
        )
        goal_sphere_urdf_location = Path("brom/models/spheres/end.urdf")
        os.makedirs(goal_sphere_urdf_location.parent, exist_ok=True)

        goal_sphere_defn.write_to_file(goal_sphere_urdf_location)

        # Load the start sphere into the plant and rigidly attach it at the start_pose
        goal_sphere_model_idx = Parser(plant).AddModels(str(goal_sphere_urdf_location))[0]
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName(f"{goal_sphere_defn.name}_base_link", goal_sphere_model_idx),
            self.goal_pose,
        )

    def add_start_and_goal_sources_to_builder(self):
        """
        Description
        -----------
        This method will add the start and goal sources to the builder.
        """
        # Input Processing
        assert self.plant.is_finalized(), "Plant must be finalized before adding start and goal sources."

        # Add the start and goal poses to the builder
        self.add_start_source_system()
        self.add_goal_source_system()

    def add_start_source_system(self):
        """
        Add a source system for the start pose.
        :return:
        """
        # Setup

        # Adds a source system for the start pose to the builder
        start_source_system = ConstantVectorSource(self.start_configuration)
        start_source_system.set_name("start_configuration")

        self.builder.AddSystem(start_source_system)

    def add_goal_source_system(self):
        """
        Add a source system for the goal pose.
        :return:
        """
        # Setup

        # Adds a source system for the goal pose to the builder
        goal_source_system = ConstantVectorSource(self.goal_configuration)
        goal_source_system.set_name("goal_configuration")

        self.builder.AddSystem(goal_source_system)

    def cast_scene(
        self,
        cast: Tuple[Role, Performer] = [],
    ):
        """
        Description
        -----------
        This small modification to the normal cast_scene()
        function includes a call to create the optional inputs if needed.
        :param cast:
        :return:
        """
        # Setup

        # Default
        super().cast_scene(cast=cast)

        # Create optional outputs, if needed
        for role_ii, performer_ii in cast:
            self.create_optional_outputs_if_necessary(
                role_ii, performer_ii
            )

    def create_optional_outputs_if_necessary(
        self,
        role: Role, # This role will always be the Motion Planner, won't it?
        performer: Performer,
    ):
        """
        Description
        -----------
        This method checks to see if all optional outputs were connected
        for the given role. If all of them were, then we do nothing.
        If any of the optional outputs were NOT connected,
        then we add a dummy value to replace it.

        Requirements
        ------------
        This method should only be called after the full scene has
        been built. Otherwise, this may have unexpected behavior.

        :return:
        """
        # Setup

        # Iterate through all OPTIONAL OUTPUTS
        # for assignment_ii in role.port_assignments:
        #     # Should we investigate this port?
        #     investigate_this_assignment = assignment_ii.pairing_type == PairingType.kOutput
        #     investigate_this_assignment = investigate_this_assignment and \
        #                                   (not assignment_ii.is_required)
        #
        #     if not investigate_this_assignment:
        #         continue # Skip this assignment value if it doesn't satisfy this
        #
        #     # Otherwise, check to see if we have the desired output port
        #     if performer.HasOutputPort(assignment_ii.performer_port_name):
        #         continue # If performer has the output port, then we should have already connected it; skip.
        #
        #     # If performer DOES NOT HAVE

        # TODO(kwesi): Maybe move above, general code to somewhere else?
        if role.name != "OfflineMotionPlanner":
            raise ValueError(
                f"Expected role for OfflineMotionPlanning scene to be \"OfflineMotionPlanner\";" +
                f"received {role.name}"
            )

        # Check to see if performer has the LAST assignment in the motion planning role
        last_assignment = role.port_assignments[-1]
        if last_assignment.performer_port_name != "plan_is_ready":
            raise ValueError(
                f"Expected last port assignment to be \"plan_is_ready\" assignment, but" +
                f"received an assignment with performer_port_name \"{last_assignment.performer_port_name}\""
            )

        if performer.HasOutputPort(last_assignment.performer_port_name):
            return # Do nothing; performer should already be connected

        # If the performer does not have plan_is_ready port, then
        # let's create a dummy value and connect it to the right place.
        plan_ready_source = ConstantValueSource(
            AbstractValue.Make(True),
        )
        self.builder.AddSystem(plan_ready_source)

        # Find system we want to connect it to
        systems_list = last_assignment.find_any_matching_input_targets(self.builder)
        assert len(systems_list) == 1, \
            f"Expected 1 system to have port \"{last_assignment.external_target_name}\"," + \
            f" but found {len(systems_list)} systems with that output port."
        # TODO(kwesi): Perhaps move more of these error assertions to a separate file?

        self.builder.Connect(
            plan_ready_source.get_output_port(),
            systems_list[0].GetInputPort(last_assignment.external_target_name)
        )

    def define_pose_ik_problem(
        self,
        pose_WorldTarget: RigidTransform,
        target_frame_name: str,
        eps0: float = 2.5e-2,
        orientation_cost_scaling: float = 0.25,
    ) -> InverseKinematics:
        """
        Description
        -----------
        Sets up the inverse kinematics problem for the start pose
        input to theis function.
        :return:
        """
        # Setup

        # Create IK Problem
        ik_problem = InverseKinematics(self.plant)

        # Add Pose Target
        ik_problem.AddPositionConstraint(
            self.plant.world_frame(),
            pose_WorldTarget.translation(),
            self.plant.GetFrameByName(target_frame_name),
            (- np.ones((3,)) * eps0).reshape((-1, 1)),
            (+ np.ones((3,)) * eps0).reshape((-1, 1)),
        )

        # TODO(kwesi): Add OrientationCosntraint
        # ik_problem.AddOrientationConstraint(
        #     self.plant.world_frame(),
        #     RotationMatrix(Quaternion(input_pose_vec[3:]).rotation()),
        #     self.plant.GetFrameByName("ft_frame"),
        #     RotationMatrix.Identity(),
        #     0.25,
        # )

        ik_problem.AddOrientationCost(
            self.plant.world_frame(),
            pose_WorldTarget.rotation(),
            self.plant.GetFrameByName(target_frame_name),
            RotationMatrix.Identity(),
            orientation_cost_scaling,
        )

        return ik_problem
    
    def easy_cast_and_build(
        self,
        planning_algorithm: Callable[
            [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
            Tuple[MotionPlan, bool],
        ],
        with_watcher: bool = True,
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        This function is used to easily cast and build the scene.
        :param planning_algorithm: The algorithm that we will use to
        plan the motion.
        :param with_watcher: A Boolean that determines whether to add a watcher to the diagram.
        :return:
        """
        # Setup

        # Add all elements to the builder
        self.add_all_secondary_cast_members_to_builder()

        # Create a planner from the algorithm
        prototypical_planner = PrototypicalPlannerSystem(
            self.plant, self.scene_graph,
            planning_algorithm,
            robot_model_idx=self.robot_model_index,
        )

        # Cast the scene using the prototypical planner
        planner_role = self.suggested_roles()[0]

        # Fulfill each role-performer pair in the casting_call list
        self.fill_role(planner_role, prototypical_planner)

        self.create_optional_outputs_if_necessary(
            planner_role, prototypical_planner
        )

        # Build
        return self.build_scene(with_watcher=with_watcher)

    @property
    def goal_configuration(self) -> np.ndarray:
        """
        Get the goal pose. This should be defined by the subclass.
        :return:
        """
        if self.goal_config_ is not None:
            return self.goal_config_
        elif self.goal_pose_ is not None:
            # Use the goal pose to get the goal configuration
            # Using the IK solver (potentially buggy because default ik problem ignores obstacles)
            return self.solve_pose_ik_problem(self.goal_pose_)
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )

    @property
    def goal_pose(self) -> RigidTransform:
        """
        Get the goal pose. This should be defined by the subclass.
        :return:
        """
        if self.goal_pose_ is not None:
            return self.goal_pose_
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )

    @property
    def robot_model_index(self) -> ModelInstanceIndex:
        """
        Get the robot model index. This should be defined by the subclass.
        :return:
        """
        if self.robot_model_idx_ is not None:
            return self.robot_model_idx_
        else:        
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )

    @property
    def start_configuration(self) -> np.ndarray:
        """
        Description:
            This function returns the start and goal poses as a single array.
        """
        if self.start_config_ is not None:
            return self.start_config_
        elif self.start_pose_ is not None:
            # Use the start pose to get the start configuration
            # Using the IK solver (potentially buggy because default ik problem ignores obstacles)
            return self.solve_pose_ik_problem(self.start_pose_)
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )
    
    @property
    def start_pose(self) -> RigidTransform:
        """
        Get the start pose. This should be defined by the subclass.
        :return:
        """
        if self.start_pose_ is not None:
            return self.start_pose_
        else:
            raise NotImplementedError(
                "This function should be implemented by the subclass."
            )

    def suggested_roles(self):
        return [kKinematicMotionPlanner]

    def solve_pose_ik_problem(
        self,
        pose_WorldTarget: RigidTransform,
        frame_name: str = "ft_frame",
    ) -> np.ndarray:
        """
        Description:
        :param input_pose_vec: An n-dimensional vector
        :return:
        """
        # Input Processing
        assert self.plant is not None, "Plant must be defined before solving IK problem!"

        # Setup

        # Define Problem
        ik_problem = self.define_pose_ik_problem(
            pose_WorldTarget,
            frame_name,
        )

        # Solve problem
        ik_program = ik_problem.prog()
        ik_result = Solve(ik_program)

        assert ik_result.get_solution_result() == SolutionResult.kSolutionFound, \
            f"Solution result was {ik_result.get_solution_result()}; need SolutionResult.kSolutionFound to make RRT Plan!"

        q_solution = ik_result.get_x_val()

        # Extract only the positions that correspond to our robot's joints
        robot_joint_names = self.plant.GetPositionNames(self.robot_model_index, add_model_instance_prefix=True)
        all_joint_names = self.plant.GetPositionNames()
        q_out_list = []
        for ii, joint_name in enumerate(all_joint_names):
            if joint_name in robot_joint_names:
                q_out_list.append(q_solution[ii])

        return np.array(q_out_list)


import numpy as np
from pydrake.all import (
    AbstractValue, BasicVector, Context,
    LeafSystem,
    PiecewisePolynomial, PiecewiseQuaternionSlerp, PiecewisePose,
    RigidTransform,
)

# Internal Imports
from brom_drake.motion_planning.systems.proximity_pose_plan_dispenser import (
    ProximityPosePlanDispenserConfig,
    DispenserTransitionRequest,
)
from brom_drake.motion_planning.systems.proximity_pose_plan_dispenser.dispenser_internal_state import (
    DispenserInternalState, DispenserTransitionRequest,
)


class ProximityPosePlanDispenser(LeafSystem):
    """
    Description
    -----------
    This class is responsible for dispensing the plan that is given to it in a way
    that is consistent with the proximity constraint that is given to it.

    Input Ports
    -----------
    plan: AbstractValue (List[RigidTransform])
        This port is used to receive the plan that the system should follow.
    plan_ready: AbstractValue (bool)
        This port is used to receive a boolean that indicates if the plan is ready.
        Lower this (to false) and then raise it (to true) to trigger the system to
        reset the plan.
    """
    def __init__(
        self,
        config: ProximityPosePlanDispenserConfig = ProximityPosePlanDispenserConfig(),
    ):
        LeafSystem.__init__(self)
        # Setup
        self.config = config

        # Create empty variables for the plan and the planned trajectory
        self.plan = []

        # Create input and output ports for the plan
        self.plan_is_set = False
        self.t0, self.t_final = -1.0, -1.0
        self.plan_port, self.request_port = None, None
        self.declare_input_ports()
        self.declare_output_ports()

        # Create internal state
        self.dispenser_state = self.DeclareDiscreteState(
            np.array([DispenserInternalState.kReady])
        )
        self.dispenser_plan_index = 0

        self.last_pose = RigidTransform()

    def advance_plan_if_necessary(self, context: Context):
        """
        Description
        -----------
        This function is responsible for advancing the plan index if the system
        is ready to do so.
        """
        # Setup
        plan = self.plan_port.Eval(context)
        plan_idx = self.dispenser_plan_index
        current_pose = self.current_pose_port.Eval(context)

        state = context.get_discrete_state(self.dispenser_state)[0]

        # If the plan is not ready, then do nothing
        if state != DispenserInternalState.kPlanSet:
            return

        # If the plan is ready, then check to see if we should advance the plan
        should_advance = self.config.in_proximity(
            current_pose,
            plan[plan_idx],
        )

        if should_advance and (plan_idx < len(plan)-1):
            self.dispenser_plan_index += 1
                    

    def declare_input_ports(self):
        """
        Description:
            This function creates the input ports for the plan dispenser.
        """
        # Setup

        # Define ports
        sample_plan = [RigidTransform()]
        self.current_pose_port = self.DeclareAbstractInputPort(
            "current_pose",
            AbstractValue.Make(RigidTransform()),
        )
        
        self.plan_port = self.DeclareAbstractInputPort(
            "plan",
            AbstractValue.Make(sample_plan),
        )

        self.request_port = self.DeclareVectorInputPort(
            "request",
            1,
        )

    def declare_output_ports(self):
        """
        Description:
            This function creates the output ports for the plan dispenser.
        """
        # Setup

        # Define ports
        self.DeclareAbstractOutputPort(
            "pose_in_plan",
            lambda: AbstractValue.Make(RigidTransform()),
            self.GetCurrentPoseInPlan,
        )

        self.DeclareVectorOutputPort(
            "dispenser_state",
            size=1,
            calc=self.GetInternalState,
        )

        self.DeclareVectorOutputPort(
            "current_pose_index",
            size=1,
            calc=self.GetCurrentPoseIndex,
        )

    def GetCurrentPoseIndex(self, context: Context, output: BasicVector):
        """
        Description
        -----------
        This function is responsible for outputting the current index of the plan that the system is on.
        """
        # Setup
        plan_idx = self.dispenser_plan_index

        # Output the current index
        output.SetFromVector(
            np.array([plan_idx])
        )

    def GetCurrentPoseInPlan(self, context: Context, output_pose: AbstractValue):
        # Setup
        plan = self.plan_port.Eval(context)
        last_pose = self.last_pose

        # Transition the internal state, if needed
        self.transition_internal_state(context)

        # Get the abstract state
        state = context.get_discrete_state(self.dispenser_state)

        # If the state of the dispenser is reset, then share
        # the LAST pose.
        dispenser_in_reset = state[0] == DispenserInternalState.kReady
        if dispenser_in_reset:
            output_pose.SetFrom(
                AbstractValue.Make(last_pose)
            )
            return
        
        # If the state of the plan is kPlanSet,
        # then check to see if we should ADVANCE to the next part of the plan
        self.advance_plan_if_necessary(context)
        plan_idx = self.dispenser_plan_index

        # Output The Current Point in the Plan
        output_pose.SetFrom(
            AbstractValue.Make(plan[plan_idx])
        )

    def GetInternalState(self, context: Context, output: BasicVector):
        """
        Description
        -----------
        Plan is set if and only if the internal variable is not None.
        """
        # Setup

        # Set the internal state (if necessary)
        self.transition_internal_state(context)

        # Output the internal state
        output.SetFrom(
            context.get_discrete_state(self.dispenser_state)
        )

    def transition_internal_state(self, context: Context):
        """
        Description
        -----------
        This function is responsible for transitioning the internal state of the system
        based on the request that is given to it.
        """
        # Setup
        current_state = context.get_discrete_state(self.dispenser_state)[0]

        # Collect the request from the port
        request = self.request_port.Eval(context)

        # Transition the state
        dispenser_is_ready = current_state == DispenserInternalState.kReady
        dispenser_plan_is_set = current_state == DispenserInternalState.kPlanSet

        if (request == DispenserTransitionRequest.kRequestSavePlan) and dispenser_is_ready:
            # Save the plan AND transition to the "Plan Set" state
            current_plan = self.plan_port.Eval(context)
            self.plan = current_plan
            self.dispenser_plan_index = 0

            # Set the internal state
            context.SetDiscreteState(
                np.array([DispenserInternalState.kPlanSet])
            )
        elif (request == DispenserTransitionRequest.kRequestReset) and dispenser_plan_is_set:
            # Prepare the system to receive a new plan
            # self.dispenser_plan_index = 0

            # Set the internal state
            context.SetDiscreteState(
                np.array([DispenserInternalState.kReady])
            )
        else:
            # Otherwise, do nothing 
            pass

    # def initialize_system_for_new_plan(self, context: Context):
    #     """
    #     Description
    #     -----------
    #     This function should lock a couple of things into memory for this object to handle future
    #     calls to get the plan.

    #     :param context:
    #     :return:
    #     """
    #     # Setup

    #     # Get time to start executing the plan
    #     self.t0 = context.get_time()
    #     context.SetDiscreteState(
    #         np.array([DispenserInternalState.kSet])
    #     )

    #     # Get times where we should reach each point in the plan and save it into a new map
    #     self.plan = self.plan_port.Eval(context)

    #     n_points_in_plan = len(self.plan)
    #     t_ii = 0.0
    #     times = [t_ii]
    #     positions_as_array = np.zeros((0, 3))
    #     for ii in range(n_points_in_plan-1):
    #         t_ii += self.find_time_between_two_points(
    #             self.plan[ii].translation(),
    #             self.plan[ii+1].translation(),
    #             )
            
    #         # Save the times, positions and orientaitons separately
    #         times += [t_ii]
    #         positions_as_array = np.vstack(
    #             (positions_as_array, self.plan[ii].translation()),
    #         )

    #     # Save final time, position and orientation
    #     self.t_final = t_ii
    #     positions_as_array = np.vstack(
    #         (positions_as_array, self.plan[-1].translation()),
    #     )

    #     # Create the planned trajectory
    #     position_trajectory = PiecewisePolynomial.FirstOrderHold(
    #         times,
    #         [p.translation() for p in self.plan],
    #     )
    #     orientation_trajectory = PiecewiseQuaternionSlerp(times, [RigidTransform(p).rotation() for p in self.plan])
        
    #     self.planned_trajectory = PiecewisePose(
    #         position_trajectory,
    #         orientation_trajectory,
    #     )





from enum import IntEnum
import numpy as np
from pydrake.all import (
    MultibodyPlant,
    RotationMatrix, RigidTransform,
    CoulombFriction,
    HalfSpace, Box,
)

# Class Definition
class GroundShape(IntEnum):
    """
    Description:
        This class defines the different shapes that the ground can have.
    """
    kHalfSpace = 0
    kBox = 1

def AddGround(
    plant: MultibodyPlant,
    shape: GroundShape = GroundShape.kBox,
):
    """
    Add a flat ground with friction
    """

    # Constants
    transparent_color = np.array([0.5, 0.5, 0.5, 0])
    nontransparent_color = np.array([0.5, 0.5, 0.5, 0.1])

    p_GroundOrigin = [0, 0.0, 0.0]
    R_GroundOrigin = RotationMatrix.MakeXRotation(0.0)
    X_GroundOrigin = RigidTransform(R_GroundOrigin,p_GroundOrigin)

    # Input Processing
    if shape == GroundShape.kHalfSpace:
        shape = HalfSpace()
    elif shape == GroundShape.kBox:
        shape = Box(10, 10, 0.1)
    else:
        raise ValueError(f"Invalid shape: {shape}!")  

    # Set Up Ground on Plant
    surface_friction = CoulombFriction(
            static_friction = 0.7,
            dynamic_friction = 0.5)
    plant.RegisterCollisionGeometry(
            plant.world_body(),
            X_GroundOrigin,
            shape,
            "ground_collision",
            surface_friction)
    plant.RegisterVisualGeometry(
            plant.world_body(),
            X_GroundOrigin,
            shape,
            "ground_visual",
            transparent_color)  # transparent


from enum import IntEnum

class DispenserInternalState(IntEnum):
    kReady = 0
    kPlanSet = 1

class DispenserTransitionRequest(IntEnum):
    kNone = 0
    kRequestSavePlan = 1
    kRequestReset = 2


"""
PortWatcher.py
Description:

    This file defines the PortWatcher class. This class is used to watch the ports of a
    diagram.
"""
from pathlib import Path
from typing import List, Tuple, Union, NamedTuple
import loguru
import numpy as np
import matplotlib.pyplot as plt
import os

from pydrake.all import (
    RigidTransform,
)
from pydrake.multibody.plant import MultibodyPlant
from pydrake.systems.framework import OutputPort, PortDataType, DiagramBuilder, LeafSystem
from pydrake.systems.primitives import LogVectorOutput
from pydrake.systems.framework import Context

# Internal Imports
from brom_drake.directories import DEFAULT_PLOT_DIR, DEFAULT_RAW_DATA_DIR
from .port_watcher_options import (
    PortWatcherOptions, FigureNamingConvention,
    PortWatcherPlottingOptions, PortWatcherRawDataOptions,
)
from .port_figure_arrangement import PortFigureArrangement
from .plotter import PortWatcherPlotter
from brom_drake.utils import RigidTransformToVectorSystem

class PortWatcher:
    def __init__(
        self,
        output_port: OutputPort,
        builder: DiagramBuilder,
        logger_name: str = None,
        options: PortWatcherOptions = PortWatcherOptions(),
        plot_dir: str = DEFAULT_PLOT_DIR,
        raw_data_dir: str = DEFAULT_RAW_DATA_DIR,
    ):
        # Setup
        self.options = options
        self.port = output_port
        self.data = {}
        self.plot_handles = {}
        self.plot_handles = None
        self.plot_dir = plot_dir
        self.raw_data_dir = raw_data_dir

        # Set up directories
        os.makedirs(self.options.plot_dir(), exist_ok=True)
        os.makedirs(self.options.raw_data_dir(), exist_ok=True)

        # Input Processing
        self.check_port_type()
        
        # Identify port's type and connect it to a logger
        system = output_port.get_system()

        if logger_name is None:
            logger_name = f"PortWatcher_{system.get_name()}_{output_port.get_name()}"

        # Preparing LogVectorSink
        self.logger = None
        self.prepare_logger(builder)
        self.logger.set_name(logger_name)

        # Prepare optional members
        self.plotter = None
        if self.options.plotting.save_to_file:
            self.plotter = PortWatcherPlotter(
                logger=self.logger,
                port=self.port,
                plotting_options=self.options.plotting,
                plot_dir=self.options.plot_dir(),
            )

    def check_port_type(self):
        """
        Description
        -----------
        Checks to see if the port is of the correct type for plotting.
        """

        # Setup
        output_port = self.port

        # Algorithm
        if output_port.get_data_type() == PortDataType.kVectorValued:
            
            return
        
        # Check to see if AbstractValue port contains RigidTransform
        output_value = output_port.Allocate()
        if isinstance(output_value.get_value(), RigidTransform):
            return

        # Raise error otherwise
        raise self.create_port_value_type_error(output_port)
    
    @staticmethod
    def create_port_value_type_error(output_port: OutputPort):
        """
        Description:
            Creates an error message for the port value type.
        :param output_port:
        :return:
        """
        # Setup
        example_value = output_port.Allocate()

        # Return
        return ValueError(
            f"This watcher only supports output ports that are:\n" +
            f"- Vector valued ports (i.e., of type {PortDataType.kVectorValued}.\n" +
            f"- Abstract valued ports containing RigidTransform objects.\n" +
            f"Received port of type {output_port.get_data_type()} with underlying type {type(example_value)}."
        )
    
    def prepare_logger(self, builder: DiagramBuilder):
        """
        Description:
            Prepares the logger for the port.
        :param builder:
        :return:
        """
        # Setup
        system = self.port.get_system()

        # Create the logger dependent on the type of data in the port
        if self.port.get_data_type() == PortDataType.kVectorValued:
            self.logger = LogVectorOutput(self.port, builder)
        else:
            # Port must be abstract valued

            # Check to see if the port contains a RigidTransform
            output_value = self.port.Allocate()
            if isinstance(output_value.get_value(), RigidTransform):
                # If it is, then we must create an intermediate system
                # that will convert the RigidTransform to a vector.
                converter_system = builder.AddSystem(
                    RigidTransformToVectorSystem()
                )

                # Connect the system to the port
                builder.Connect(
                    self.port,
                    converter_system.get_input_port(),
                )
                # Then connect the output of the converter to a logger
                self.logger = LogVectorOutput(
                    converter_system.get_output_port(),
                    builder,
                )
            else:
                raise NotImplementedError(
                    f"PortWatcher does not support the type of data contained in the port."
                )
    
    def safe_system_name(self) -> str:
        """
        Description:
            Returns a safe name for the system.
        :param name: System's name.
        :return:
        """
        # Setup
        system = self.port.get_system()
        out = system.get_name()

        # First, let's check to see how many "/" exist in the name
        slash_occurences = [i for i, letter in enumerate(out) if letter == "/"]
        if len(slash_occurences) > 0:
            out = out[slash_occurences[-1] + 1:]  # truncrate string based on the last slash

        # Second, replace all spaces with underscores
        out = out.replace(" ", "_")

        return out

    def save_raw_data(self, diagram_context: Context):
        """
        Description:
            Saves the raw data to a file.
        :param diagram_context:
        :return:
        """
        # Setup
        log = self.logger.FindLog(diagram_context)
        time_data_file_name, raw_data_file_name = self.time_and_raw_data_names()

        # Save time data
        log_times = log.sample_times()
        os.makedirs(Path(time_data_file_name).parent, exist_ok=True)
        np.save(time_data_file_name, log_times)

        # Save the data
        log_data = log.data()
        np.save(raw_data_file_name, log_data)

    def time_and_raw_data_names(self) -> Tuple[str, str]:
        """
        Description:
            Returns the names that will be given to the raw data for this port.
        
        :return: Tuple of strings where:
        - the first string is the name of the time data and,
        - the second string is the name of the data.
        """
        # Setup
        options = self.options
        format = options.raw_data.file_format
        raw_data_dir = options.raw_data_dir()

        # If this has the flat naming convention, then the file should be contained within the plot_dir.
        return [
            f"{raw_data_dir}/system_{self.safe_system_name()}_port_{self.port.get_name()}_times.{format}",
            f"{raw_data_dir}/system_{self.safe_system_name()}_port_{self.port.get_name()}.{format}"
        ]



import networkx as nx
import numpy as np
from pydrake.systems.framework import LeafSystem, Diagram
from pydrake.systems.primitives import VectorLogSink
from typing import Union

Performer = Union[LeafSystem, Diagram]
MotionPlan = Union[nx.DiGraph, np.ndarray]
SupportedLogger = Union[VectorLogSink]



DEFAULT_BROM_DIR = "./brom"
DEFAULT_WATCHER_DIR = DEFAULT_BROM_DIR + "/watcher"
DEFAULT_PLOT_DIR = DEFAULT_WATCHER_DIR + "/plots"
DEFAULT_RAW_DATA_DIR = DEFAULT_WATCHER_DIR + "/raw_data"
DEFAULT_BROM_MODELS_DIR = DEFAULT_BROM_DIR + "/models"



from importlib import resources as impresources
import networkx as nx
import numpy as np
from pydrake.all import (
    MultibodyPlant, Parser,
    CollisionFilterDeclaration, GeometrySet,
)
from pydrake.common.eigen_geometry import Quaternion
from pydrake.math import RollPitchYaw, RigidTransform
from pydrake.systems.framework import Diagram, Context
from typing import Callable, Tuple

from brom_drake.directories import DEFAULT_BROM_MODELS_DIR
from brom_drake.file_manipulation.urdf import drakeify_my_urdf
from brom_drake.file_manipulation.urdf.shapes.box import BoxDefinition
from brom_drake.file_manipulation.urdf.simple_writer.urdf_definition import SimpleShapeURDFDefinition, \
    InertiaDefinition
from brom_drake.motion_planning.systems.open_loop_dispensers.open_loop_plan_dispenser import OpenLoopPlanDispenser
import brom_drake.robots as robots
from brom_drake.robots.stations.kinematic import UR10eStation as KinematicUR10eStation
from brom_drake.scenes import SceneID
from brom_drake.scenes.roles import Role
from brom_drake.scenes.types import KinematicMotionPlanningScene
from brom_drake.utils import Performer, AddGround, MotionPlan


class ChemLab1Scene(KinematicMotionPlanningScene):
    """
    Description:
        This scene is the first in the chemistry lab series.
    """
    def __init__(
        self,
        time_step=1e-3,
        meshcat_port_number: int = 7000,
        plan_execution_speed: float = 0.2,
        table_length: float = 0.6,
        table_width: float = 2.0,
        table_height: float = 0.1,
        shelf_pose: RigidTransform = None,
        **kwargs,
    ):
        """
        Description:
            Constructor for the ChemLab1Scene class.
        :param meshcat_port_number:
        """
        # Superclass constructor
        super().__init__(**kwargs)

        # Input Processing
        self.time_step = time_step
        self.meshcat_port_number = meshcat_port_number
        self.plan_execution_speed = plan_execution_speed
        self.table_height, self.table_width = table_height, table_width
        self.table_length = table_length

        # Set shelf pose
        self.shelf_pose = shelf_pose
        if self.shelf_pose is None:
            self.pose_WorldShelf = RigidTransform(
                RollPitchYaw(0.0, 0.0, -np.pi/2.0).ToQuaternion(),
                np.array([0.0, 0.75, 0.66]),
            )

        # Set Beaker Pose
        self.pose_WorldBeaker = RigidTransform(
            RollPitchYaw(np.pi/2.0, 0.0, 0.0).ToQuaternion(),
            np.array([-0.6, 0.45, 0.075]),
        )

        # Set Holder Pose
        self.pose_WorldHolder = RigidTransform(
            RollPitchYaw(np.pi/2.0, 0.0, 0.0).ToQuaternion(),
            np.array([self.table_width*0.5*0.7, 0.6+self.table_length/4., self.table_height-0.015]),
        )

        # Set station
        self.station = KinematicUR10eStation(
            time_step=self.time_step,
            meshcat_port_number=self.meshcat_port_number,
        )
        self.arm = self.station.arm
        self.robot_model_idx_ = self.arm

        # Set Names of Plant and scene graph
        self.plant = self.station.plant
        self.scene_graph = self.station.scene_graph

        self.plant.set_name(f"ChemLabScene1_Plant")
        self.scene_graph.set_name(f"ChemLabScene1_SceneGraph")

        # Define placeholder variables for models
        self.test_tube_holder1 = None

    def add_all_secondary_cast_members_to_builder(self):
        """
        Description
        -----------
        This method adds all secondary cast members to the builder.
        The secondary cast members in the scene are the:
        - Table, where the robot exists
        - Test Tube Holders
        - ...
        :return:
        """
        # Call the superclass method
        super().add_all_secondary_cast_members_to_builder()

        # Setup

        # Create the table and the items on it
        self.add_table()
        self.add_test_tube_holders()
        self.add_beaker()
        self.add_shelf()
        AddGround(self.plant)

        # Add the station
        self.builder.AddSystem(
            self.station,
        )

        self.station.Finalize()

        # Add the motion planning components
        self.add_robot_source_system()
        self.add_motion_planning_components()
        self.add_start_and_goal_sources_to_builder()

    def add_beaker(self):
        """
        Description
        -----------
        This method adds the beaker to the scene.
        """
        # Setup
        plant = self.plant
        urdf_file_path = str(
            impresources.files(robots) / "models/beaker/beaker.urdf"
        )

        # Use Drakeify my urdf to create the beaker
        new_beaker_urdf = drakeify_my_urdf(urdf_file_path)

        # Add the beaker to the plant
        model_idcs = Parser(plant=plant).AddModels(str(new_beaker_urdf))
        self.beaker_model_index = model_idcs[0]

        # Weld the beaker to the world frame
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("beaker_base_link", self.beaker_model_index),
            self.pose_WorldBeaker,
        )

    def add_motion_planning_components(self):
        """
        Add the motion planning components to the builder.
        :return:
        """
        # Setup
        n_actuated_dof = self.plant.num_actuated_dofs()

        # Add the Plan Dispenser and connect it to the station
        self.plan_dispenser = self.builder.AddSystem(
            OpenLoopPlanDispenser(n_actuated_dof, self.plan_execution_speed)
        )

        self.builder.Connect(
            self.plan_dispenser.GetOutputPort("point_in_plan"),
            self.station.GetInputPort("desired_joint_positions"),
        )

    def add_shelf(self):
        """
        Add the shelf to the scene.
        :return:
        """
        # Setup
        plant = self.plant
        urdf_file_path = str(
            impresources.files(robots) / "models/cupboard/cupboard.sdf"
        )

        # Add the shelf to the plant
        model_idcs = Parser(plant=plant).AddModels(urdf_file_path)
        self.shelf_model_index = model_idcs[0]

        # Weld the bookshelf to the world frame
        plant.WeldFrames(
            plant.world_frame(),
            plant.GetFrameByName("cupboard_body", self.shelf_model_index),
            self.pose_WorldShelf,
        )

    def add_table(self):
        """
        Description
        -----------
        This method adds the table to the scene.
        The table will be a simple shape that the robot will interact with.
        :return:
        """
        # Setup

        # Create a new shape urdf
        table_defn = SimpleShapeURDFDefinition(
            name="table",
            shape=BoxDefinition(
                size=(self.table_width, self.table_length, self.table_height),
            ),
            mass=100.0, # kg
            inertia=InertiaDefinition(
                ixx=10.0,
                iyy=10.0,
                izz=10.0,
            ),
            color=np.array([0.1, 0.1, 0.1, 0.5]),
        )
        table_urdf_path = DEFAULT_BROM_MODELS_DIR + "/table/table.urdf"
        table_defn.write_to_file(table_urdf_path)

        # Add the table to the scene
        table_model_index = Parser(self.plant).AddModels(table_urdf_path)[0]

        # Weld the table to the world frame
        table_pose = RigidTransform(
            RollPitchYaw(0.0, 0.0, 0.0).ToQuaternion(),
            np.array([0.0, 0.6, 0.0]),
        )
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName("table_base_link", table_model_index),
            table_pose,
        )

    def add_test_tube_holders(self):
        """
        Description
        -----------
        This method adds the test tube holders to the scene.
        :return:
        """
        # Setup
        test_tube_urdf1 = str(
            impresources.files(robots) / "models/test_tube_holder/test_tube_holder.urdf"
        )

        # Use Drakeify my urdf to create the test tube holders
        new_test_tube_holder_urdf1 = drakeify_my_urdf(test_tube_urdf1)

        self.test_tube_holder1 = Parser(self.plant).AddModels(
            str(new_test_tube_holder_urdf1)
        )[0]

        # Weld the test tube holders to the world frame
        self.plant.WeldFrames(
            self.plant.world_frame(),
            self.plant.GetFrameByName("test_tube_holder_base_link", self.test_tube_holder1),
            self.pose_WorldHolder,
        )


    def cast_scene_and_build(
        self,
        cast: Tuple[Role, Performer] = [],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        Modifies the normal cast_scene_and_build, so that
        we share the context of the plant with the appropriate
        parts of the system.
        :param cast:
        :param with_watcher: A Boolean that determines whether to add a watcher to the diagram.
        :return:
        """
        diagram, diagram_context = super().cast_scene_and_build(
            cast=cast,
            with_watcher=with_watcher,
        )

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        for role_ii, performer_ii in cast:
            if role_ii.name == "OfflineMotionPlanner":
                performer_ii.set_internal_root_context(
                    diagram_context
                )

        return diagram, diagram_context
    
    def configure_collision_filter(self, scene_graph_context: Context):
        """
        Description
        -----------
        This method configures the collision filter for the scene.
        :param scene_graph_context:
        :return:
        """
        # Setup
        scene_graph = self.scene_graph
        shelf_model_index = self.shelf_model_index

        # Ignore self collisions of the shelf, by:
        # - Getting the model instance index of the shelf
        # - Collecting the Geometry IDs of all the geometries in the shelf
        shelf_geometry_ids = []
        for body_index in self.plant.GetBodyIndices(shelf_model_index):
            # Get the geometry IDs
            shelf_geometry_ids.extend(
                self.plant.GetCollisionGeometriesForBody(
                    self.plant.get_body(body_index)
                )
            )
            # print(self.plant.get_body(body_index))

        self.shelf_geometry_ids = shelf_geometry_ids

        # Apply the collision filter
        scene_graph.collision_filter_manager(scene_graph_context).Apply(
            CollisionFilterDeclaration().ExcludeWithin(
                GeometrySet(self.shelf_geometry_ids)
            )
        )

        # Ignore collisions between the robot links
        # TODO: Actually implement this for adjacent links? Or is this not necessary?
        arm_geometry_ids = []
        for body_index in self.plant.GetBodyIndices(self.arm):
            arm_geometry_ids.extend(
                self.plant.GetCollisionGeometriesForBody(
                    self.plant.get_body(body_index)
                )
            )

        self.arm_geometry_ids = arm_geometry_ids

        # Apply the collision filter
        scene_graph.collision_filter_manager(scene_graph_context).Apply(
            CollisionFilterDeclaration().ExcludeWithin(
                GeometrySet(self.arm_geometry_ids)
            )
        )
    
    def easy_cast_and_build(
        self,
        planning_algorithm: Callable[
            [np.ndarray, np.ndarray, Callable[[np.ndarray], bool]],
            Tuple[MotionPlan, bool],
        ],
        with_watcher: bool = False,
    ) -> Tuple[Diagram, Context]:
        """
        Description
        -----------
        This function is used to easily cast and build the scene.
        :param planning_algorithm: The algorithm that we will use to
        plan the motion.
        :param with_watcher: A Boolean that determines whether to add a watcher to the diagram.
        :return:
        """
        # Setup

        # Use Base class implementation to start
        diagram, diagram_context = super().easy_cast_and_build(
            planning_algorithm,
            with_watcher=with_watcher,
        )

        # Configure the scene graph for collision detection
        self.configure_collision_filter(
            diagram.GetSubsystemContext(
                self.scene_graph, diagram_context,
            )
        )

        # Connect arm controller to the appropriate plant_context
        self.station.arm_controller.plant_context = diagram.GetSubsystemContext(
            self.station.arm_controller.plant, diagram_context,
        )

        self.performers[0].set_internal_root_context(
            diagram_context
        )

        return diagram, diagram_context

    @property
    def goal_configuration(self):
        """
        Get the goal pose. This should be defined by the subclass.
        :return:
        """
        if self.goal_config_ is None:
            beaker_to_goal_translation = np.array([+0.0, 0.2, 0.0]) 
            beaker_to_goal_orientation = RollPitchYaw(0., 0., 0.0).ToQuaternion()
            X_BeakerGoal = RigidTransform(  
                beaker_to_goal_orientation,
                beaker_to_goal_translation,
            )

            pose_WorldGoal = self.pose_WorldBeaker.multiply(X_BeakerGoal)

            # Use Inverse Kinematics to get the goal configuration of the robot
            self.goal_config_ = self.solve_pose_ik_problem(
                pose_WorldGoal,
            )

            return self.goal_config_
        else:
            return self.goal_config_

    @property
    def goal_pose(self) -> RigidTransform:
        """
        Description
        -----------
        Get the goal pose. This should be defined by the subclass.
        """
        # Setup

        # Algorithm
        if self.goal_pose_ is None:
            beaker_to_goal_translation = np.array([+0.0, 0.2, 0.0]) 
            beaker_to_goal_orientation = RollPitchYaw(0., 0., 0.0).ToQuaternion()
            X_BeakerGoal = RigidTransform(  
                beaker_to_goal_orientation,
                beaker_to_goal_translation,
            )

            pose_WorldGoal = self.pose_WorldBeaker.multiply(X_BeakerGoal)
            self.goal_pose_ = pose_WorldGoal

            return self.goal_pose_
        else:
            return self.goal_pose_

    @property
    def id(self) -> SceneID:
        return SceneID.kShelfPlanning1
        
    # TODO(kwesi): Implement start_configuration method.

    @property
    def start_pose(self) -> RigidTransform:
        """
        Get the start pose. This should be defined by the subclass.
        :return:
        """
        # Setup

        if self.start_pose_ is None:
            # Define Start Pose
            holder_to_start_translation = np.array([+0.0, 0.2, 0.025])
            holder_to_start_orientation = Quaternion(1, 0, 0, 0)
            X_HolderStart = RigidTransform(
                holder_to_start_orientation,
                holder_to_start_translation,
            )
            pose_WorldStart = self.pose_WorldHolder.multiply(X_HolderStart)
            self.start_pose_ = pose_WorldStart
            return self.start_pose_
        else:
            return self.start_pose_



from dataclasses import dataclass

@dataclass
class InertiaDefinition:
    """
    A dataclass that defines the inertia of a simple shape.
    """
    ixx: float = 1.0
    ixy: float = 0.0
    ixz: float = 0.0
    iyy: float = 1.0
    iyz: float = 0.0
    izz: float = 1.0

    def as_list(self) -> list:
        """
        Convert the inertia to a list.
        :return: A list of the inertia values.
        """
        return [self.ixx, self.ixy, self.ixz, self.iyy, self.iyz, self.izz]

    def as_map(self) -> dict:
        """
        Convert the inertia to a map.
        :return: A map of the inertia values.
        """
        return {
            "ixx": f"{self.ixx}",
            "ixy": f"{self.ixy}",
            "ixz": f"{self.ixz}",
            "iyy": f"{self.iyy}",
            "iyz": f"{self.iyz}",
            "izz": f"{self.izz}",
        }


from dataclasses import dataclass
from enum import IntEnum
import numpy as np
from typing import Union
import xml.etree.ElementTree as ET


class ShapeEnum(IntEnum):
    """
    The different simple shapes that can be created.
    """
    kBox = 1
    kSphere = 2
    kCylinder = 3
    kCapsule = 4

# Shape Definition Can be any one of these
@dataclass
class ShapeDefinition:
    def add_geometry_to_element(self, target_element: ET.Element):
        raise NotImplementedError("This method must be implemented in the subclass.")

    @property
    def type(self):
        raise NotImplementedError("This method must be implemented in the subclass.")



import numpy as np
from pydrake.all import (
    AbstractValue, BasicVector,
    Context,
    LeafSystem,
    RigidTransform,
)

class RigidTransformToVectorSystem(LeafSystem):
    def __init__(self):
        LeafSystem.__init__(self)
        self.rigid_transform_input = self.DeclareAbstractInputPort(
            "rigid_transform",
            AbstractValue.Make(RigidTransform()),
        )
        self.DeclareVectorOutputPort(
            "vector_xyz_quat(wxyz)",
            7,
            self.CalcVectorOutput,
        )

    def CalcVectorOutput(self, context: Context, output: BasicVector):
        rigid_transform = self.rigid_transform_input.Eval(context)
        xyz = rigid_transform.translation()
        quat = rigid_transform.rotation().ToQuaternion().wxyz()
        output.SetFromVector(
            np.hstack((xyz, quat))
        )


from enum import Enum, unique

@unique
class SceneID(Enum):
    kNotDefined = "NotDefined"
    kShowMeThisModel = "ShowMeThisModel"
    kShelfPlanning1 = "ShelfPlanning1"
    kChemLab1 = "ChemLab-Planning1"


from dataclasses import dataclass
import numpy as np
from typing import Union
import xml.etree.ElementTree as ET

# Internal Import
from .shape_definition import ShapeDefinition, ShapeEnum

@dataclass
class BoxDefinition(ShapeDefinition):
    size: Union[float, list, tuple, np.ndarray]

    def add_geometry_to_element(self, target_element: ET.Element):
        # Setup

        # Convert the size to a numpy array
        if isinstance(self.size, (list, tuple, np.ndarray)):
            self.size = np.array(self.size)
        elif isinstance(self.size, float):
            self.size = np.array([self.size, self.size, self.size])
        else:
            raise ValueError(f"Size {self.size} (type \"{type(self.size)}\") not supported.")

        # Return the element
        return ET.SubElement(
            target_element,
            "box",
            {"size": f"{self.size[0]} {self.size[1]} {self.size[2]}"},
        )

    @property
    def type(self):
        return ShapeEnum.kBox


from .search import (
    find_all_systems_with_input_port,
    find_all_systems_with_output_port,
)
from .constants import Performer, MotionPlan
from .ground import AddGround, GroundShape
from .rigid_transform_to_vector_system import RigidTransformToVectorSystem

__all__ = [
    "find_all_systems_with_input_port",
    "find_all_systems_with_output_port",
    "Performer",
    "MotionPlan",
    "AddGround",
    "GroundShape",
    "RigidTransformToVectorSystem",
]


import loguru
import matplotlib.pyplot as plt
import numpy as np
import os
from pathlib import Path
from pydrake.all import (
    Context,
    LeafSystem,
    OutputPort,
    PortDataType,
    MultibodyPlant,
    RigidTransform,
)
from typing import List, Tuple


# Internal Imports
from brom_drake.PortWatcher.port_figure_arrangement import PortFigureArrangement
from brom_drake.PortWatcher.port_watcher_options import FigureNamingConvention, PortWatcherPlottingOptions
from brom_drake.utils.constants import SupportedLogger
from brom_drake.directories import DEFAULT_PLOT_DIR

class PortWatcherPlotter:
    def __init__(
        self,
        logger: SupportedLogger,
        port: OutputPort,
        plotting_options: PortWatcherPlottingOptions = PortWatcherPlottingOptions(),
        plot_dir: str = DEFAULT_PLOT_DIR,
    ):
        # Setup
        self.logger = logger
        self.port = port
        self.plotting_options = plotting_options
        self.plot_dir = plot_dir

    def compute_plot_shape(self, n_dims: int) -> Tuple[int, int]:
        """
        Description:
            Computes the shape of the plot based on the data.
        :param n_dims: The number of dimensions in the data.
        :return:
        """
        if n_dims == 1:
            return 1, 1
        if n_dims == 2:
            return 1, 2

        if n_dims < 9:
            return 2, int(np.ceil(n_dims / 2.0))

        # Otherwise
        return 3, int(np.ceil(n_dims / 3.0))

    def figure_names(self) -> List[str]:
        """
        Description:
            Returns the name of the figure.
        :return:
        """
        # Setup
        plotting_options = self.plotting_options

        # If this has the flat naming convention, then the file should be contained within the plot_dir.
        if plotting_options.figure_naming_convention == FigureNamingConvention.kFlat:
            return self.figure_names_under_flat_convention()


        elif plotting_options.figure_naming_convention == FigureNamingConvention.kHierarchical:
            return self.figure_names_under_hierarchical_convention()

        else:
            raise NotImplementedError(
                f"Invalid figure naming convention for figure_names(): {plotting_options.figure_naming_convention}."
            )


    def figure_names_under_flat_convention(self) -> List[str]:
        """
        Description:
            Returns the names associated with each figure that this port will
            generate assuming we are under the kFlat convention.
        :param self:
        :return: List of strings where each string is a file name for an associated figure.
        """
        # Setup
        plotting_options = self.plotting_options
        format = plotting_options.file_format
        plot_dir = self.plot_dir

        #
        if plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            return [
                f"{plot_dir}/system_{self.safe_system_name()}_port_{self.port.get_name()}.{format}"
            ]

        elif plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerDim:
            return [
                f"{plot_dir}/system_{self.safe_system_name()}_port_{self.port.get_name()}_dim{ii}.{format}"
                for ii in range(self.port.size())
            ]

        else:
            raise NotImplementedError(
                f"Invalid plot arrangement for figure naming convention {options.plotting.figure_naming_convention}: {options.plot_arrangement}."
            )

    def figure_names_under_hierarchical_convention(self) -> List[str]:
        """
        Description:
            Returns the names associated with each figure that this port will
            generate assuming we are under the kHierarchical convention.
        :return:
        """
        # Setup
        plotting_options = self.plotting_options
        format = plotting_options.file_format
        plot_dir = self.plot_dir

        # Algorithm
        if plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            return [
                f"{plot_dir}/system_{self.safe_system_name()}/port_{self.port.get_name()}.{format}"
            ]

        elif plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerDim:
            return [
                f"{plot_dir}/system_{self.safe_system_name()}/port_{self.port.get_name()}/{self.name_of_data_at_index(ii, remove_spaces=True)}.{format}"
                for ii in range(self.port.size())
            ]

        else:
            raise NotImplementedError(
                f"Invalid plot arrangement for figure naming convention {plotting_options.figure_naming_convention}: {options.plot_arrangement}."
            )

    def name_of_data_at_index(
        self,
        dim_index: int,
        remove_spaces: bool = False,
    ) -> str:
        """
        Description:
            Returns the name of the data which is in index dim_index
            of this vector-valued port.
        :param dim_index:
        :param remove_spaces: Whether to remove spaces from the name.
        :return:
        """
        # Setup
        plotting_options = self.plotting_options
        n_dims = self.data_dimension()
        system = self.port.get_system()

        # Input Processing
        if dim_index >= n_dims:
            raise ValueError(
                f"dim_index ({dim_index}) is greater than the number of dimensions in the port ({n_dims})."
            )

        # Default name
        name = f"Dim #{dim_index}"

        # Only use full names, if we are NOT using the OnePlotPerPort config
        if plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            return name

        if self.system_is_multibody_plant():
            # The multi-body plant has names for specific ports
            if self.port.get_name() == "state":
                # We can get the names of the state
                state_names = system.GetStateNames()
                name = state_names[dim_index]
            else:
                # Announce that we are using the default name
                loguru.logger.warning(
                    f"Using default name for data at index {dim_index} of port {self.port.get_name()} of system {system.get_name()}."
                )

        # Filter our spaces, if requested
        if remove_spaces:
            name = name.replace(" ", "_")

        # Return name!
        return name

    def data_dimension(self) -> int:
        """
        Description:
            Returns the dimension of the data in the port.
        :return:
        """
        if self.port.get_data_type() == PortDataType.kVectorValued:
            return self.port.size()
        else:
            # If port contains RigidTransform, then the expected data dimension is 7.
            example_value = self.port.Allocate()
            if isinstance(example_value.get_value(), RigidTransform):
                return 7
        
        # Otherwise, raise an error
        raise ValueError(
            f"Port {self.port.get_name()} of system {self.port.get_system().get_name()} is not of the correct type for plotting."
        )

    def plot_logger_data(
        self,
        diagram_context: Context,
    ) -> Tuple[List[plt.Figure], List[List[plt.Axes]]]:
        """
        plot_logger_data
        Description:

            This function plots the data in the logger.
        :param diagram_context:
        :return:
        """
        # Setup
        logger = self.logger
        plotting_options = self.plotting_options
        system = self.port.get_system()

        # Get the log from the logger
        temp_log = logger.FindLog(diagram_context)

        log_times = temp_log.sample_times()
        log_data = temp_log.data()

        if (log_data.shape[1] == 0) or (log_data.shape[0] == 0):
            loguru.logger.warning(
                f"No data found for {system.get_name()} - Port {self.port.get_name()}! Skipping...")
            return None, None

        # Plot the data
        if plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerPort:
            fig, ax_list = self.plot_logger_data_subplots(log_times, log_data)
            return [fig], [ax_list]

        elif plotting_options.plot_arrangement == PortFigureArrangement.OnePlotPerDim:
            figs, ax_grid = [], []
            for port_index in range(self.port.size()):
                fig_ii = plt.figure()
                ax_ii = fig_ii.add_subplot(1, 1, 1)
                ax_ii.plot(
                    log_times, log_data[port_index, :]
                )

                # Add axis titles and labels
                ax_ii.set_xlabel("Time (s)")
                ax_ii.set_title(self.name_of_data_at_index(port_index))

                # Save figures and axes to lists
                figs.append(fig_ii)
                ax_grid.append([ax_ii])

            return figs, ax_grid

        else:
            raise ValueError(
                f"Invalid plot arrangement: {plotting_options.plot_arrangement}."
            )

    def plot_logger_data_subplots(
        self,
        times: np.array,
        data: np.array,

    ):
        """
        plot_logger_data_subplots
        Description:

            This function plots the data in the logger.
        :param times:
        :param data:
        :param diagram_context:
        :return:
        """
        # Setup
        n_dims = data.shape[0]

        # Plot the data
        n_rows, n_cols = self.compute_plot_shape(n_dims)

        print(f"Plotting {n_dims} dimensions in a {n_rows}x{n_cols} grid.")

        fig, ax_list = plt.subplots(n_rows, n_cols)

        if n_rows == 1 and n_cols == 1:
            ax_list.plot(times, data[0, :])
            ax_list.set_title(self.name_of_data_at_index(0))

        elif (n_rows == 1) or (n_cols == 1):
            for dim_index in range(n_dims):
                ax_list[dim_index].plot(times, data[dim_index, :])
                ax_list[dim_index].set_title(
                    self.name_of_data_at_index(dim_index),
                )

        else:
            for row_index in range(n_rows):
                for col_index in range(n_cols):

                    dim_index = n_cols * row_index + col_index

                    if dim_index >= n_dims:
                        fig.delaxes(ax_list[row_index, col_index])
                        continue

                    ax_list[row_index, col_index].plot(times, data[dim_index, :])
                    ax_list[row_index, col_index].set_title(self.name_of_data_at_index(dim_index))

        return fig, ax_list
    
    def safe_system_name(self) -> str:
        """
        Description:
            Returns a safe name for the system.
        :param name: System's name.
        :return:
        """
        # Setup
        system = self.port.get_system()
        out = system.get_name()

        # First, let's check to see how many "/" exist in the name
        slash_occurences = [i for i, letter in enumerate(out) if letter == "/"]
        if len(slash_occurences) > 0:
            out = out[slash_occurences[-1] + 1:]  # truncrate string based on the last slash

        # Second, replace all spaces with underscores
        out = out.replace(" ", "_")

        return out

    def save_figures(self, diagram_context: Context):
        """
        save_figures
        Description:
            This function saves the figures.
        TODO(kwesi): Make it so that this function computes names + directory structure based on plot arrangement.
        :param diagram_context:
        :return:
        """
        # Setup
        plotting_options = self.plotting_options
        figs, ax_list = self.plot_logger_data(diagram_context)

        # If no figures are returned, then return early!
        if figs is None:
            return

        if len(figs) == 0:
            return # Do nothing

        # Save the figures
        figure_names = self.figure_names()

        if len(figs) == 1:
            # Create directory for the plots, if it doesn't already exist
            os.makedirs(Path(figure_names[0]).parent, exist_ok=True)
            figs[0].savefig(figure_names[0], dpi=plotting_options.plot_dpi)
            plt.close(figs[0])
        else:
            # Plot each figure within this directory
            for ii, fig_ii in enumerate(figs):
                os.makedirs(Path(figure_names[ii]).parent, exist_ok=True)
                fig_ii.savefig(figure_names[ii],dpi=plotting_options.plot_dpi)
                plt.close(fig_ii)

        # Close all figures when done (this should be redundant?)
        # plt.close('all')

    def system_is_multibody_plant(self) -> bool:
        """
        Description:
            Returns True if the system is a MultibodyPlant.
        :return:
        """
        # Setup
        system = self.port.get_system()

        # Return
        return type(system) == MultibodyPlant


from dataclasses import dataclass
import xml.etree.ElementTree as ET

# Internal Import
from .shape_definition import ShapeDefinition, ShapeEnum

@dataclass
class SphereDefinition(ShapeDefinition):
    radius: float

    def add_geometry_to_element(self, target_element: ET.Element):
        return ET.SubElement(
            target_element,
            "sphere",
            {"radius": f"{self.radius}"},
        )

    @property
    def type(self):
        return ShapeEnum.kSphere